#!/usr/bin/env python

from io import BytesIO
from struct import pack, unpack_from
import math

class Table(BytesIO):
    _formats = {
        'i8': 'b', 'u8': 'B',
        'i16': '>h', 'u16': '>H',
        'i32': '>i', 'u32': '>I',
        'i64': '>q'
    }

    def __init__(self, tag=None):
        self.tag = tag

    def __getattr__(self, name):
        if name in self._formats:
            return lambda val: self.write(pack(self._formats[name], val))
        raise AttributeError

    def add_padding(self):
        size = self.tell()
        self.write(bytes(((size + 3) & ~3) - size))

    def calc_checksum(self):
        view = self.getbuffer()
        val = 0
        for i in range(0, view.nbytes, 4):
            val = (val + unpack_from('>I', view, i)[0]) & 0xffffffff
        return val

class Glyph:
    def __init__(self):
        self.bitmap = []

    def get_bitaligned_bitmap(self):
        linepad = (8 - self.bbx_w) % 8
        v = 0
        for line in self.bitmap:
            v = (v << self.bbx_w) | (line >> linepad)
        pad = (8 - self.bbx_w * self.bbx_h) % 8
        v <<= pad
        size = (self.bbx_w * self.bbx_h + 7) >> 3
        return v.to_bytes(size, 'big')

class Strike:
    def __init__(self):
        self.w = 0
        self.h = 0
        self.glyphs = {}

    def set_codes_and_sizes(self):
        self.codes = sorted(self.glyphs.keys())
        self.sizes = []
        cur_bbx_w = None
        cur_bbx_h = None
        for i, code in enumerate(self.codes):
            glyph_id = i + 1
            glyph = self.glyphs[code]
            if glyph.bbx_w != cur_bbx_w or glyph.bbx_h != cur_bbx_h:
                if self.sizes:
                    self.sizes[-1].append(glyph_id - 1)
                self.sizes.append([glyph.bbx_w, glyph.bbx_h, glyph_id])
                cur_bbx_w = glyph.bbx_w
                cur_bbx_h = glyph.bbx_h
        self.sizes[-1].append(glyph_id)

def make_unicode_maps():
    maps = {}

    maps['ISO8859-1'] = {
        i: i for i in (*range(0x20, 0x7f), *range(0xa0, 0x100))
    }

    maps['JISX0201.1976-0'] = {
        **{i: i for i in range(0x20, 0x7f)},
        **{
            0x5c: 0xa5, # YEN SIGN
            0x7e: 0x203e, # OVERLINE
        },
        **{i: i + 0xfec0 for i in range(0xa1, 0xe0)},
    }

    d = {}
    for i in range(0x2120, 0x7e7f):
        try:
            d[i] = ord((i | 0x8080).to_bytes(2, 'big').decode('euc_jis_2004')[0])
        except UnicodeDecodeError:
            pass
    d.update({
        0x215d: 0xff0d, # FULLWIDTH HYPHEN-MINUS
        0x2171: 0xffe0, # FULLWIDTH CENT SIGN
        0x2172: 0xffe1, # FULLWIDTH POUND SIGN
        0x224c: 0xffe2, # FULLWIDTH NOT SIGN
        0x2925: 0xffe4, # FULLWIDTH BROKEN BAR
    })
    maps['JISX0213-1.2004-0'] = d

    maps['JISX0208.1990-0'] = maps['JISX0213-1.2004-0']

    return maps

def load_bdf(strikes, unicode_maps, f, verbose):
    inbitmap = False
    for line in f:
        args = line.split()
        if args[0] == 'SIZE':
            strike = strikes.setdefault(int(args[1]), Strike())
        elif args[0] == 'FONTBOUNDINGBOX':
            strike.w = max(strike.w, int(args[1]))
            strike.h = max(strike.h, int(args[2]))
        elif args[0] == 'CHARSET_REGISTRY':
            charset_registry = args[1].strip('"')
        elif args[0] == 'CHARSET_ENCODING':
            charset_encoding = args[1].strip('"')
        elif args[0] == 'ENDPROPERTIES':
            unicode_map = unicode_maps[charset_registry + '-' + charset_encoding]
        elif args[0] == 'STARTCHAR':
            glyph = Glyph()
        elif args[0] == 'ENDCHAR':
            if verbose and code in strike.glyphs:
                print(f'{f.name}: override {encoding:04x}({encoding})'
                      f' ({code:04x}({code}))', file=sys.stderr)
            strike.glyphs[code] = glyph;
            inbitmap = False
        elif args[0] == 'DWIDTH':
            glyph.dwidth = int(args[1])
        elif args[0] == 'BBX':
            glyph.bbx_w = int(args[1])
            glyph.bbx_h = int(args[2])
        elif args[0] == 'ENCODING':
            encoding = int(args[1])
            code = unicode_map[encoding]
        elif args[0] ==  'BITMAP':
            inbitmap = True
        elif inbitmap:
            glyph.bitmap.append(int(line, 16))

def calc_segs(codes):
    segs = []
    i = 1
    start = end = codes[0]
    starti = i;
    idrangeofs = 0
    for code in codes[1:]:
        i += 1
        if code == end + 1:
            end = code
        else:
            iddelta = -start + starti
            segs.append((start, end, iddelta, idrangeofs))
            start = end = code
            starti = i
    iddelta = -start + starti
    segs.append((start, end, iddelta, idrangeofs))
    segs.append((0xffff, 0xffff, 1, 0))
    return segs

def build_ttf(f, strikes, fontname, vender_id , copyright, revision):
    strike1 = strikes[max(strikes.keys())]
    nglyphs = len(strike1.codes)
    em = 1024
    ascender = em * strike1.h // strike1.w

    eblc = Table('EBLC')
    eblc.u16(2) # majorVersion
    eblc.u16(0) # minorVersion
    eblc.u32(len(strikes)) # numSizes

    ebdt = Table('EBDT')
    ebdt.u16(2) # majorVersion
    ebdt.u16(0) # minorVersion

    eblc_header_size = 8
    bitmapSizeTable_size = 4 * 4 + 12 + 12 + 8
    indexSubTableArray_size = 8
    indexSubTable_size = 20
    indexSubTableArrayOffset = eblc_header_size + bitmapSizeTable_size * len(strikes)
    subtab = Table()
    for size, strike in strikes.items():
        ## subtables and EBDT bitmap
        subtabarr_start = subtab.tell()
        ofs = indexSubTableArray_size * len(strike.sizes)
        for w, h, first, last in strike.sizes:
            # indexSubTableArray
            subtab.u16(first) # firstGlyphIndex
            subtab.u16(last) # lastGlyphIndex
            subtab.u32(ofs) # additionalOffsetToIndexSubtable
            ofs += indexSubTable_size

        subtab_start = subtab.tell()
        for w, h, first, last in strike.sizes:
            # indexSubTables
            #   header
            subtab.u16(2) # indexFormat
            subtab.u16(5) # imageFormat
            subtab.u32(ebdt.tell()) # imageDataOffset
            subtab.u32((w * h + 7) >> 3) # imageSize
            #   bigMetrics
            subtab.u8(h) # height
            subtab.u8(w) # width
            subtab.i8(0) # horiBearingX
            subtab.i8(h) # horiBearingY
            subtab.u8(w) # horiAdvance
            subtab.i8(-w // 2) # vertBearingX
            subtab.i8(0) # vertBearingY
            subtab.u8(h) # vertAdvance
            # EBDT format 5
            for code in strike.codes[first-1:last]:
                ebdt.write(strike.glyphs[code].get_bitaligned_bitmap())

        ## bitmap size table
        indexSubTableArrayOffset += subtabarr_start
        eblc.u32(indexSubTableArrayOffset) # indexSubTableArrayOffset
        eblc.u32(subtab.tell() - subtab_start) # indexTablesSize
        eblc.u32(len(strike.sizes)) # numberOfIndexSubTables
        eblc.u32(0) # colorRef (not used)

        # SbitLineMetrics hori
        eblc.i8(strike.w) # ascender
        eblc.i8(strike.h) # descender
        eblc.u8(size) # widthMax
        eblc.i8(1) # caretSlopeNumerator
        eblc.i8(0) # caretSlopeDenominator
        eblc.i8(0) # caretOffset
        eblc.i8(0) # minOriginSB
        eblc.i8(size) # minAdvanceSB
        eblc.i8(strike.w) # maxBeforeBL
        eblc.i8(strike.h) # minAfterBL
        eblc.i8(0) # pad1
        eblc.i8(0) # pad2

        # SbitLineMetrics vert
        eblc.i8(0) # ascender
        eblc.i8(0) # descender
        eblc.u8(0) # widthMax
        eblc.i8(0) # caretSlopeNumerator
        eblc.i8(0) # caretSlopeDenominator
        eblc.i8(0) # caretOffset
        eblc.i8(0) # minOriginSB
        eblc.i8(0) # minAdvanceSB
        eblc.i8(0) # maxBeforeBL
        eblc.i8(0) # minAfterBL
        eblc.i8(0) # pad1
        eblc.i8(0) # pad2

        eblc.u16(1) # startGlyphIndex
        eblc.u16(nglyphs) # endGlyphIndex
        eblc.u8(size) # ppemX
        eblc.u8(size) # ppemY
        eblc.u8(1) # bitDepth
        eblc.i8(1) # flags (1: horizontal)

    # indexSubTableArray and indexSubTables
    eblc.write(subtab.getbuffer())

    glyf = Table('glyf')
    glyf.i16(1) # numberOfContours
    glyf.i16(0) # xMin
    glyf.i16(0) # yMin
    glyf.i16(0) # xMax
    glyf.i16(0) # yMax
    glyf.i16(0) # endPtsOfContours[numberOfContours]
    glyf.i16(0) # instuctionLength
    glyf.u8(0) # flags
    glyf.i16(0) # xCoordinates
    glyf.i16(0) # yCoordinates

    loca = Table('loca')
    # offsets[n]
    loca.u32(0) # for glyph ID 0
    for i in range(nglyphs):
        loca.u32(0)
    loca.u32(glyf.tell()) # for end

    cmap = Table('cmap')
    cmap.u16(0) # version
    cmap.u16(1) # numTables
    # encodingRecords[numTables]
    cmap.u16(3) # platformID (3: Windows)
    cmap.u16(1) # encodingID (1: Unicode BMP)
    cmap.u32(cmap.tell() + 4) # offset (4: size of Offset32)
    # Subtable Format 4
    segs = calc_segs(strike1.codes)
    segcount = len(segs)
    cmap.u16(4) # format
    cmap.u16(16 + 8 * segcount) # length
    cmap.u16(0) # language
    cmap.u16(2 * segcount) # segCountX2
    searchrange = 2 * (2 ** math.floor(math.log2(segcount)))
    cmap.u16(searchrange) # searchRange
    cmap.u16(int(math.log2(searchrange / 2))) # entrySelector
    cmap.u16(2 * segcount - searchrange) # rangeShift
    for start, end, iddelta, idrangeofs in segs:
        cmap.u16(end) # endCode
    cmap.u16(0) # reservedPad
    for start, end, iddelta, idrangeofs in segs:
        cmap.u16(start) # startCode
    for start, end, iddelta, idrangeofs in segs:
        cmap.u16(iddelta & 0xffff) # idDelta
    for start, end, iddelta, idrangeofs in segs:
        cmap.u16(idrangeofs) # idRangeOffset[segCount]

    os2 = Table('OS/2')
    os2.u16(1) # version
    os2.i16(em // 2) # xAvgCharWidth
    os2.u16(400) # usWeightClass (400: Normal (Regular))
    os2.u16(5) # usWidthClass (5: Medium (normal))
    os2.u16(8) # fsType (8: Editable embedding)
    os2.i16(em // 2) # ySubscriptXSize
    os2.i16(em // 2) # ySubscriptYSize
    os2.i16(0) # ySubscriptXOffset
    os2.i16(0) # ySubscriptYOffset
    os2.i16(em // 2) # ySuperscriptXSize
    os2.i16(em // 2) # ySuperscriptYSize
    os2.i16(0) # ySuperscriptXOffset
    os2.i16(em // 2) # ySuperscriptYOffset
    os2.i16(em // 20) # yStrikeoutSize
    os2.i16(260) # yStrikeoutPosition
    os2.i16(0) # sFamilyClass
    os2.u8(2) # panose bFamilyType (2: Latin Text)
    os2.u8(0) # panose bSerifStyle (0: Any)
    os2.u8(5) # panose bWeight (5: Book)
    os2.u8(9) # panose bProportion (9: Monospaced)
    os2.u8(0) # panose bContrast (0: Any)
    os2.u8(0) # panose bStrokeVariation (0: Any)
    os2.u8(0) # panose bArmStyle (0: Any)
    os2.u8(0) # panose bLetterform (0: Any)
    os2.u8(0) # panose bMidline (0: Any)
    os2.u8(0) # panose bXHeight (0: Any)
    os2.u32(0xffffffff) # ulUnicodeRange1 XXX
    os2.u32(0xffffffff) # ulUnicodeRange2 XXX
    os2.u32(0x0000003f) # ulUnicodeRange3 XXX
    os2.u32(0x00000000) # ulUnicodeRange4
    os2.write(vender_id.encode()) # achVendID
    os2.u16(0x40) # fsType (bit 6: REGULAR)
    os2.u16(min(strike1.codes)) # usFirstCharIndex
    os2.u16(max(strike1.codes)) # usLastCharIndex
    os2.i16(ascender) # sTypoAscender
    os2.i16(0) # sTypoDescender
    os2.i16(0) # sTypoLineGap
    os2.u16(ascender) # usWinAscent
    os2.u16(0) # usWinDescent
    os2.u32(0x00020001) # ulCodePageRange1 (Latin 1, JIS/Japan)
    os2.u32(0x00000000) # ulCodePageRange2

    head = Table('head')
    head.u16(1) # majorVersion
    head.u16(0) # minorVersion
    revparts = tuple(int(x) for x in revision.split('.'))
    if len(revparts) == 2:
        rev = revparts[0] << 16 | revparts[1] << 8
    else:
        rev = revparts[0] << 16 | revparts[1] << 8 | revparts[2]
    head.i32(rev) # fontRevision (not used by Windows)
    head.u32(0) # checkSumAdjustment (set after calculate sum)
    head.u32(0x5f0f3cf5) # magicNumber
    head.u16(0xb) # flags
    head.u16(em) # unitsPerEm
    head.i64(0) # created (0 for idempotence)
    head.i64(0) # modified (0 for idempotence)
    head.i16(0) # xMin
    head.i16(0) # yMin
    head.i16(em) # xMax
    head.i16(ascender) # yMax
    head.u16(0) # macStyle
    head.u16(7) # lowestRecPPEM
    head.i16(2) # fontDirectionHint
    head.i16(1) # indexToLocFormat (1: long offsets)
    head.i16(0) # glyphDataFormat

    hhea = Table('hhea')
    hhea.u16(1) # majorVersion
    hhea.u16(0) # minorVersion
    hhea.i16(ascender) # ascender
    hhea.i16(0) # descender
    hhea.i16(0) # lineGap
    hhea.u16(em) # advanceWidthMax
    hhea.u16(0) # minLeftSideBearing
    hhea.u16(0) # minRightSideBearing
    hhea.u16(em) # xMaxExtent
    hhea.i16(1) # caretSlopeRise (1: vertical)
    hhea.i16(0) # caretSlopeRun
    hhea.i16(0) # caretOffset
    hhea.i16(0) # (reserved)
    hhea.i16(0) # (reserved)
    hhea.i16(0) # (reserved)
    hhea.i16(0) # (reserved)
    hhea.i16(0) # metricDataFormat
    hhea.u16(nglyphs + 1) # numberOfHMetrics

    hmtx = Table('hmtx')
    # hMetrics[numberOfHMetrics]
    #  glyph ID 0
    hmtx.u16(em // 2) # advanceWidth
    hmtx.i16(0) # lsb
    for code in strike1.codes:
        glyph = strike1.glyphs[code]
        hmtx.u16(em * glyph.dwidth // strike1.w) # advanceWidth
        hmtx.i16(0) # lsb

    maxp = Table('maxp')
    maxp.i32(0x00010000) # version
    maxp.u16(nglyphs + 1) # numGlyphs
    maxp.u16(1) # maxPoints
    maxp.u16(1) # maxContours
    maxp.u16(0) # maxCompositePoints
    maxp.u16(0) # maxCompositeContours
    maxp.u16(1) # maxZones
    maxp.u16(0) # maxTwilightPoints
    maxp.u16(0) # maxStorage
    maxp.u16(0) # maxFunctionDefs
    maxp.u16(0) # maxInstructionDefs
    maxp.u16(0) # maxStackElements
    maxp.u16(0) # maxSizeOfInstructions
    maxp.u16(0) # maxComponentElements
    maxp.u16(0) # maxComponentDepth

    post = Table('post')
    post.i32(0x00030000) # version
    post.i32(0) # italicAngle
    post.u16(0) # underlinePosition
    post.u16(1) # underlineThickness
    post.u32(1) # isFixedPitch
    post.u32(0) # minMemType42
    post.u32(0) # maxMemType42
    post.u32(0) # minMemType1
    post.u32(0) # maxMemType1

    name = Table('name')
    names = (
        copyright, fontname, 'Regular', fontname, fontname,
        'Version ' + revision
    )
    langids = (0x0409, 0x0411) # English US, Japanese
    names = [s.encode('utf-16be') for s in names]
    count = len(names) * len(langids)
    nameHeader_size = 6
    nameRecord_size = 12
    name.u16(0) # format
    name.u16(count) # count
    name.u16(nameHeader_size + nameRecord_size * count) # stringOffset
    # nameRecord[count]
    ofs = 0
    for langid in langids:
        for i, s in enumerate(names):
            name.u16(3) # platformID (3: Windows)
            name.u16(1) # encodingID (1: Unicode BMP)
            name.u16(langid) # languageID
            name.u16(i) # nameID
            name.u16(len(s)) # length
            name.u16(ofs) # offset
            ofs += len(s)
    for langid in langids:
        for s in names:
            name.write(s)

    ## output

    # must be sorted by table tag
    tables = (
        ebdt, eblc, os2, cmap, glyf, head, hhea, hmtx, loca, maxp, name, post,
    )
    ntables = len(tables)

    ttf = Table()

    # offset table
    search_range = 128 # (maximum power of 2 <= numTables) * 16
    ttf.u32(0x00010000) # sfntVersion
    ttf.u16(ntables) # numTables
    ttf.u16(search_range) # searchRange
    ttf.u16(3) # entrySelector (log2(maximum power of 2 <= numTables))
    ttf.u16(ntables * 16 - search_range) # rangeShift

    # table records
    table_record_size = 16
    offset = ttf.tell() + table_record_size * ntables
    checksumadj = 0
    for table in tables:
        ttf.write(table.tag.encode()) # tableTag
        origlen = table.tell()
        table.add_padding()
        checksum = table.calc_checksum()
        checksumadj += checksum
        ttf.u32(checksum) # checkSum
        ttf.u32(offset) # offset
        ttf.u32(origlen) # length
        offset += table.tell()
    checksumadj = (0xb1b0afba - checksumadj) & 0xffffffff
    head.seek(8)
    head.u32(checksumadj)

    f.write(ttf.getbuffer())
    for table in tables:
        f.write(table.getbuffer())

if __name__ == '__main__':
    from getopt import getopt
    import sys

    verbose = False
    opts, args = getopt(sys.argv[1:], 'v')
    for o, a in opts:
        if o == '-v':
            verbose = True

    fontname = args[0]
    vender_id = args[1]
    copyright = args[2]
    revision = args[3]
    ttf = args[4]
    bdfs = args[5:]

    strikes = {}
    unicode_maps = make_unicode_maps()
    for bdf in bdfs:
        load_bdf(strikes, unicode_maps, open(bdf), verbose)
    for strike in strikes.values():
        strike.set_codes_and_sizes()

    with open(ttf, 'wb') as f:
        build_ttf(f, strikes, fontname, vender_id, copyright, revision)
